@@ -42,12 +42,11 @@
 #include <util/util.h>
 #include <infiniband/verbs.h>
 #include <ccan/array_size.h>
-#include <ccan/bitmap.h>
+#include <util/bitmap.h>
 #include <ccan/container_of.h>
 #include <linux/if_ether.h>
 #include "hns_roce_u_abi.h"
 
-#define HNS_ROCE_HW_VER1		('h' << 24 | 'i' << 16 | '0' << 8 | '6')
 #define HNS_ROCE_HW_VER2		0x100
 #define HNS_ROCE_HW_VER3		0x130
 
@@ -59,9 +58,7 @@
 
 #define HNS_ROCE_MAX_RC_INL_INN_SZ	32
 #define HNS_ROCE_MAX_UD_INL_INN_SZ	8
-#define HNS_ROCE_MAX_CQ_NUM		0x10000
 #define HNS_ROCE_MIN_CQE_NUM		0x40
-#define HNS_ROCE_V1_MIN_WQE_NUM		0x20
 #define HNS_ROCE_V2_MIN_WQE_NUM		0x40
 #define HNS_ROCE_MIN_SRQ_WQE_NUM	1
 
@@ -75,9 +72,6 @@
 
 #define HNS_ROCE_GID_SIZE		16
 
-#define HNS_ROCE_CQ_DB_BUF_SIZE		((HNS_ROCE_MAX_CQ_NUM >> 11) << 12)
-#define HNS_ROCE_STATIC_RATE		3 /* Gbps */
-
 #define INVALID_SGE_LENGTH 0x80000000
 
 #define HNS_ROCE_DWQE_PAGE_SIZE 65536
@@ -159,13 +153,6 @@
 #define HNS_ROCE_SRQ_TABLE_BITS 8
 #define HNS_ROCE_SRQ_TABLE_SIZE BIT(HNS_ROCE_SRQ_TABLE_BITS)
 
-/* operation type list */
-enum {
-	/* rq&srq operation */
-	HNS_ROCE_OPCODE_SEND_DATA_RECEIVE         = 0x06,
-	HNS_ROCE_OPCODE_RDMA_WITH_IMM_RECEIVE     = 0x07,
-};
-
 struct hns_roce_device {
 	struct verbs_device		ibv_dev;
 	int				page_size;
@@ -188,12 +175,18 @@ enum hns_roce_db_type {
 	HNS_ROCE_DB_TYPE_NUM
 };
 
+enum hns_roce_pktype {
+	HNS_ROCE_PKTYPE_ROCE_V1,
+	HNS_ROCE_PKTYPE_ROCE_V2_IPV6,
+	HNS_ROCE_PKTYPE_ROCE_V2_IPV4,
+};
+
 struct hns_roce_db_page {
 	struct hns_roce_db_page	*prev, *next;
 	struct hns_roce_buf	buf;
 	unsigned int		num_db;
 	unsigned int		use_cnt;
-	bitmap			*bitmap;
+	unsigned long		*bitmap;
 };
 
 struct hns_roce_context {
@@ -201,8 +194,6 @@ struct hns_roce_context {
 	void				*uar;
 	pthread_spinlock_t		uar_lock;
 
-	void				*cq_tptr_base;
-
 	struct {
 		struct hns_roce_qp	**table;
 		int			refcnt;
@@ -228,6 +219,8 @@ struct hns_roce_context {
 	unsigned int			max_srq_sge;
 	int				max_cqe;
 	unsigned int			cqe_size;
+	uint32_t			config;
+	unsigned int			max_inline_data;
 };
 
 struct hns_roce_pd {
@@ -236,7 +229,7 @@ struct hns_roce_pd {
 };
 
 struct hns_roce_cq {
-	struct ibv_cq			ibv_cq;
+	struct verbs_cq			verbs_cq;
 	struct hns_roce_buf		buf;
 	pthread_spinlock_t		lock;
 	unsigned int			cqn;
@@ -247,6 +240,7 @@ struct hns_roce_cq {
 	int				arm_sn;
 	unsigned long			flags;
 	unsigned int			cqe_size;
+	struct hns_roce_v2_cqe		*cqe;
 };
 
 struct hns_roce_idx_que {
@@ -258,10 +252,21 @@ struct hns_roce_idx_que {
 	unsigned int			tail;
 };
 
+struct hns_roce_rinl_wqe {
+	struct ibv_sge			*sg_list;
+	unsigned int			sge_cnt;
+};
+
+struct hns_roce_rinl_buf {
+	struct hns_roce_rinl_wqe	*wqe_list;
+	unsigned int			wqe_cnt;
+};
+
 struct hns_roce_srq {
 	struct verbs_srq		verbs_srq;
 	struct hns_roce_idx_que		idx_que;
 	struct hns_roce_buf		wqe_buf;
+	struct hns_roce_rinl_buf	srq_rinl_buf;
 	pthread_spinlock_t		lock;
 	unsigned long			*wrid;
 	unsigned int			srqn;
@@ -281,6 +286,7 @@ struct hns_roce_wq {
 	unsigned int			head;
 	unsigned int			tail;
 	unsigned int			max_gs;
+	unsigned int			ext_sge_cnt;
 	unsigned int			rsv_sge;
 	unsigned int			wqe_shift;
 	unsigned int			shift; /* wq size is 2^shift */
@@ -301,21 +307,6 @@ struct hns_roce_sge_ex {
 	unsigned int			sge_shift;
 };
 
-struct hns_roce_rinl_sge {
-	void				*addr;
-	unsigned int			len;
-};
-
-struct hns_roce_rinl_wqe {
-	struct hns_roce_rinl_sge	*sg_list;
-	unsigned int			sge_cnt;
-};
-
-struct hns_roce_rinl_buf {
-	struct hns_roce_rinl_wqe	*wqe_list;
-	unsigned int			wqe_cnt;
-};
-
 struct hns_roce_qp {
 	struct verbs_qp			verbs_qp;
 	struct hns_roce_buf		buf;
@@ -337,6 +328,12 @@ struct hns_roce_qp {
 	unsigned long			flags;
 	int				refcnt; /* specially used for XRC */
 	void				*dwqe_page;
+
+	/* specific fields for the new post send APIs */
+	int				err;
+	void				*cur_wqe;
+	unsigned int			rb_sq_head; /* roll back sq head */
+	struct hns_roce_sge_info	sge_info;
 };
 
 struct hns_roce_av {
@@ -406,7 +403,7 @@ static inline struct hns_roce_pd *to_hr_pd(struct ibv_pd *ibv_pd)
 
 static inline struct hns_roce_cq *to_hr_cq(struct ibv_cq *ibv_cq)
 {
-	return container_of(ibv_cq, struct hns_roce_cq, ibv_cq);
+	return container_of(ibv_cq, struct hns_roce_cq, verbs_cq.cq);
 }
 
 static inline struct hns_roce_srq *to_hr_srq(struct ibv_srq *ibv_srq)
@@ -447,6 +444,8 @@ int hns_roce_u_bind_mw(struct ibv_qp *qp, struct ibv_mw *mw,
 struct ibv_cq *hns_roce_u_create_cq(struct ibv_context *context, int cqe,
 				    struct ibv_comp_channel *channel,
 				    int comp_vector);
+struct ibv_cq_ex *hns_roce_u_create_cq_ex(struct ibv_context *context,
+					  struct ibv_cq_init_attr_ex *cq_attr);
 
 int hns_roce_u_modify_cq(struct ibv_cq *cq, struct ibv_modify_cq_attr *attr);
 int hns_roce_u_destroy_cq(struct ibv_cq *cq);
@@ -493,7 +492,6 @@ void hns_roce_free_qp_buf(struct hns_roce_qp *qp, struct hns_roce_context *ctx);
 
 void hns_roce_init_qp_indices(struct hns_roce_qp *qp);
 
-extern const struct hns_roce_u_hw hns_roce_u_hw_v1;
 extern const struct hns_roce_u_hw hns_roce_u_hw_v2;
 
 #endif /* _HNS_ROCE_U_H */
