@@ -82,6 +82,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_FLOW_METER]	= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_MISS]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_DECAP] = {
 			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
@@ -99,6 +100,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_MODIFY_HDR] = {
 			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
@@ -114,6 +116,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_POP_VLAN] = {
 			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
@@ -130,6 +133,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_PUSH_VLAN] = {
 			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
@@ -143,6 +147,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_DEST_ARRAY]	= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_NON_TERM] = {
 			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
@@ -164,6 +169,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ASO] = {
 			[DR_ACTION_TYP_QP]              = DR_ACTION_STATE_TERM,
@@ -174,6 +180,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_FIRST_HIT]   = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]          = DR_ACTION_STATE_ASO,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ENCAP] = {
 			[DR_ACTION_TYP_QP]		= DR_ACTION_STATE_TERM,
@@ -184,6 +191,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_TAG]		= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_ASO_FIRST_HIT]	= DR_ACTION_STATE_ASO,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_TERM] = {
 			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
@@ -204,12 +212,14 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_MISS]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ENCAP] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_ASO_FIRST_HIT]	= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_MODIFY_HDR] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -219,6 +229,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_PUSH_VLAN]	= DR_ACTION_STATE_PUSH_VLAN,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_POP_VLAN] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -230,6 +241,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_MODIFY_HDR]	= DR_ACTION_STATE_MODIFY_HDR,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_PUSH_VLAN] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -239,6 +251,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_PUSH_VLAN]	= DR_ACTION_STATE_PUSH_VLAN,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]	= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_NON_TERM] = {
 			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
@@ -254,6 +267,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_MISS]		= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ASO] = {
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
@@ -267,6 +281,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_DROP]            = DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_FT]              = DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_MISS]            = DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_TERM] = {
 			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
@@ -292,6 +307,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_MISS]		= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_DECAP] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -308,6 +324,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ENCAP] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -317,6 +334,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_METER]		= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_ASO_FIRST_HIT]	= DR_ACTION_STATE_ASO,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_MODIFY_HDR] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -331,6 +349,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_POP_VLAN] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -346,6 +365,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_PUSH_VLAN] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -359,6 +379,8 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 		},
 		[DR_ACTION_STATE_NON_TERM] = {
 			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
@@ -379,6 +401,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_L2_TO_TNL_L3]    = DR_ACTION_STATE_ENCAP,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ASO] = {
 			[DR_ACTION_TYP_VPORT]           = DR_ACTION_STATE_TERM,
@@ -389,6 +412,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_FIRST_HIT]   = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]          = DR_ACTION_STATE_ASO,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_TERM] = {
 			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
@@ -412,6 +436,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]		= DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_MISS]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ENCAP] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -421,6 +446,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_SAMPLER]		= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_DEST_ARRAY]	= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_MODIFY_HDR] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -433,6 +459,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_L2_TO_TNL_L3]	= DR_ACTION_STATE_ENCAP,
 			[DR_ACTION_TYP_PUSH_VLAN]	= DR_ACTION_STATE_PUSH_VLAN,
 			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_POP_VLAN] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -447,6 +474,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_SAMPLER]         = DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_DEST_ARRAY]      = DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_VPORT]           = DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_PUSH_VLAN] = {
 			[DR_ACTION_TYP_FT]		= DR_ACTION_STATE_TERM,
@@ -459,6 +487,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_SAMPLER]		= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_DEST_ARRAY]	= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_VPORT]		= DR_ACTION_STATE_TERM,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_NON_TERM] = {
 			[DR_ACTION_TYP_DROP]		= DR_ACTION_STATE_TERM,
@@ -477,6 +506,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_MISS]		= DR_ACTION_STATE_TERM,
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]          = DR_ACTION_STATE_ASO,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_ASO] = {
 			[DR_ACTION_TYP_L2_TO_TNL_L2]    = DR_ACTION_STATE_ENCAP,
@@ -491,6 +521,7 @@ static const enum dr_action_valid_state next_action_state[DR_ACTION_DOMAIN_MAX]
 			[DR_ACTION_TYP_ASO_FIRST_HIT]   = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_FLOW_METER]  = DR_ACTION_STATE_ASO,
 			[DR_ACTION_TYP_ASO_CT]          = DR_ACTION_STATE_ASO,
+			[DR_ACTION_TYP_ROOT_FT]		= DR_ACTION_STATE_TERM,
 		},
 		[DR_ACTION_STATE_TERM] = {
 			[DR_ACTION_TYP_CTR]		= DR_ACTION_STATE_TERM,
@@ -632,17 +663,31 @@ int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 			attr.final_icm_addr = nic_dmn->drop_icm_addr;
 			break;
 		case DR_ACTION_TYP_FT:
-			if (action->dest_tbl->dmn != dmn) {
+		{
+			struct mlx5dv_dr_table *dest_tbl = action->dest_tbl;
+
+			if (dest_tbl->dmn != dmn) {
 				dr_dbg(dmn, "Destination table belongs to a different domain\n");
 				goto out_invalid_arg;
 			}
-			if (action->dest_tbl->level <= matcher->tbl->level) {
+			if (dest_tbl->level <= matcher->tbl->level) {
 				dr_dbg(dmn, "Destination table level should be higher than source table\n");
 				goto out_invalid_arg;
 			}
 			attr.final_icm_addr = rx_rule ?
-				action->dest_tbl->rx.s_anchor->chunk->icm_addr :
-				action->dest_tbl->tx.s_anchor->chunk->icm_addr;
+				dr_icm_pool_get_chunk_icm_addr(dest_tbl->rx.s_anchor->chunk) :
+				dr_icm_pool_get_chunk_icm_addr(dest_tbl->tx.s_anchor->chunk);
+			break;
+		}
+		case DR_ACTION_TYP_ROOT_FT:
+			if (action->root_tbl.tbl->dmn != dmn) {
+				dr_dbg(dmn, "Destination anchor belongs to a different domain\n");
+				goto out_invalid_arg;
+
+			}
+			attr.final_icm_addr = rx_rule ?
+				action->root_tbl.rx_icm_addr :
+				action->root_tbl.tx_icm_addr;
 			break;
 		case DR_ACTION_TYP_QP:
 			if (action->dest_qp.is_qp)
@@ -702,18 +747,45 @@ int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 				dr_dbg(dmn, "Root decap L3 action cannot be used on current table\n");
 				goto out_invalid_arg;
 			}
-			attr.decap_index = action->rewrite.index;
-			attr.decap_actions = action->rewrite.num_of_actions;
-			attr.decap_with_vlan =
-				attr.decap_actions == WITH_VLAN_NUM_HW_ACTIONS;
+			if (action->rewrite.ptrn_arg.ptrn && action->rewrite.ptrn_arg.arg) {
+				attr.decap_index =
+					dr_arg_get_object_id(action->rewrite.ptrn_arg.arg);
+				attr.decap_actions =
+				     action->rewrite.ptrn_arg.ptrn->rewrite_param.num_of_actions;
+				attr.decap_pat_idx =
+					action->rewrite.ptrn_arg.ptrn->rewrite_param.index;
+			} else {
+				attr.decap_index = action->rewrite.param.index;
+				attr.decap_actions = action->rewrite.param.num_of_actions;
+				attr.decap_with_vlan =
+					attr.decap_actions == WITH_VLAN_NUM_HW_ACTIONS;
+				attr.decap_pat_idx = DR_INVALID_PATTERN_INDEX;
+			}
 			break;
 		case DR_ACTION_TYP_MODIFY_HDR:
 			if (action->rewrite.is_root_level) {
 				dr_dbg(dmn, "Root modify header action cannot be used on current table\n");
 				goto out_invalid_arg;
 			}
-			attr.modify_index = action->rewrite.index;
-			attr.modify_actions = action->rewrite.num_of_actions;
+
+			if (action->rewrite.single_action_opt) {
+				attr.modify_actions = action->rewrite.param.num_of_actions;
+				attr.single_modify_action = action->rewrite.param.data;
+			} else {
+				if (action->rewrite.ptrn_arg.ptrn && action->rewrite.ptrn_arg.arg) {
+					attr.modify_index =
+						dr_arg_get_object_id(action->rewrite.ptrn_arg.arg);
+					attr.modify_pat_idx =
+						action->rewrite.ptrn_arg.ptrn->rewrite_param.index;
+					attr.modify_actions =
+						action->rewrite.ptrn_arg.ptrn->rewrite_param.
+										num_of_actions;
+				} else {
+					attr.modify_actions = action->rewrite.param.num_of_actions;
+					attr.modify_index = action->rewrite.param.index;
+					attr.modify_pat_idx = DR_INVALID_PATTERN_INDEX;
+				}
+			}
 			break;
 		case DR_ACTION_TYP_L2_TO_TNL_L2:
 		case DR_ACTION_TYP_L2_TO_TNL_L3:
@@ -770,15 +842,9 @@ int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 				goto out_invalid_arg;
 			}
 			attr.hit_gvmi = action->vport.caps->vhca_gvmi;
-			if (rx_rule) {
-				/* Loopback on WIRE vport is not supported */
-				if (action->vport.caps->num == WIRE_PORT)
-					goto out_invalid_arg;
-
-				attr.final_icm_addr = action->vport.caps->icm_address_rx;
-			} else {
-				attr.final_icm_addr = action->vport.caps->icm_address_tx;
-			}
+			attr.final_icm_addr = rx_rule ?
+				action->vport.caps->icm_address_rx :
+				action->vport.caps->icm_address_tx;
 			break;
 		case DR_ACTION_TYP_DEST_ARRAY:
 			if (action->dest_array.dmn != dmn) {
@@ -798,7 +864,7 @@ int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 			}
 
 			max_actions_type = MAX_VLANS;
-			attr.vlans.count++;
+			attr.vlans.count_pop++;
 			break;
 		case DR_ACTION_TYP_PUSH_VLAN:
 			if (rx_rule && !(dmn->ste_ctx->actions_caps &
@@ -808,12 +874,12 @@ int dr_actions_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
 			}
 
 			max_actions_type = MAX_VLANS;
-			if (attr.vlans.count == MAX_VLANS) {
+			if (attr.vlans.count_push == MAX_VLANS) {
 				errno = ENOTSUP;
 				return ENOTSUP;
 			}
 
-			attr.vlans.headers[attr.vlans.count++] = action->push_vlan.vlan_hdr;
+			attr.vlans.headers[attr.vlans.count_push++] = action->push_vlan.vlan_hdr;
 			break;
 		default:
 			goto out_invalid_arg;
@@ -1031,6 +1097,112 @@ dec_ref:
 	return NULL;
 }
 
+static int dr_action_create_dest_root_table(struct mlx5dv_dr_action *action)
+{
+	struct mlx5dv_dr_domain *dmn = action->root_tbl.tbl->dmn;
+	struct dr_devx_flow_dest_info dest_info = {};
+	struct dr_devx_flow_table_attr ft_attr = {};
+	struct dr_devx_flow_group_attr fg_attr = {};
+	struct dr_devx_flow_fte_attr fte_attr = {};
+	int ret;
+
+	switch (dmn->type) {
+	case MLX5DV_DR_DOMAIN_TYPE_FDB:
+		ft_attr.type = FS_FT_FDB;
+		break;
+	case MLX5DV_DR_DOMAIN_TYPE_NIC_RX:
+		ft_attr.type = FS_FT_NIC_RX;
+		break;
+	case MLX5DV_DR_DOMAIN_TYPE_NIC_TX:
+		ft_attr.type = FS_FT_NIC_TX;
+		break;
+	default:
+		errno = EOPNOTSUPP;
+		return errno;
+	}
+
+	fte_attr.dest_arr = &dest_info;
+
+	fte_attr.action |= MLX5_FLOW_CONTEXT_ACTION_FWD_DEST;
+	fte_attr.dest_arr[0].type = MLX5_FLOW_DEST_TYPE_FT;
+	fte_attr.dest_arr[0].ft_id = action->root_tbl.sa->id;
+	fte_attr.dest_size = 1;
+
+	action->root_tbl.devx_tbl = dr_devx_create_always_hit_ft(dmn->ctx,
+							       &ft_attr,
+							       &fg_attr,
+							       &fte_attr);
+	if (!action->root_tbl.devx_tbl)
+		return errno;
+
+	ret = dr_devx_query_flow_table(action->root_tbl.devx_tbl->ft_dvo,
+				       ft_attr.type,
+				       &action->root_tbl.rx_icm_addr,
+				       &action->root_tbl.tx_icm_addr);
+	if (ret)
+		goto destroy_devx_tbl;
+
+	return 0;
+
+destroy_devx_tbl:
+	dr_devx_destroy_always_hit_ft(action->root_tbl.devx_tbl);
+
+	return errno;
+}
+
+struct mlx5dv_dr_action *
+mlx5dv_dr_action_create_dest_root_table(struct mlx5dv_dr_table *tbl,
+					uint16_t priority)
+{
+	struct mlx5dv_steering_anchor_attr attr = {};
+	enum mlx5_ib_uapi_flow_table_type ft_type;
+	struct mlx5dv_steering_anchor *sa;
+	struct mlx5dv_dr_action *action;
+	int err;
+
+	if (!dr_is_root_table(tbl)) {
+		dr_dbg(tbl->dmn, "Supported only on root flow tables\n");
+		errno = EINVAL;
+		return NULL;
+	}
+
+	if (tbl->dmn->type == MLX5DV_DR_DOMAIN_TYPE_NIC_RX)
+		ft_type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_RX;
+	else if (tbl->dmn->type == MLX5DV_DR_DOMAIN_TYPE_NIC_TX)
+		ft_type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_NIC_TX;
+	else
+		ft_type = MLX5_IB_UAPI_FLOW_TABLE_TYPE_FDB;
+
+	attr.priority = priority;
+	attr.ft_type = ft_type;
+
+	sa = mlx5dv_create_steering_anchor(tbl->dmn->ctx, &attr);
+	if (!sa)
+		return NULL;
+
+	action = dr_action_create_generic(DR_ACTION_TYP_ROOT_FT);
+	if (!action)
+		goto free_steering_anchor;
+
+	action->root_tbl.sa = sa;
+	action->root_tbl.tbl = tbl;
+
+	err = dr_action_create_dest_root_table(action);
+	if (err)
+		goto free_action;
+
+	atomic_fetch_add(&tbl->refcount, 1);
+
+	return action;
+
+free_action:
+	free(action);
+free_steering_anchor:
+	mlx5dv_destroy_steering_anchor(sa);
+
+	return NULL;
+}
+
 struct mlx5dv_dr_action *
 mlx5dv_dr_action_create_flow_counter(struct mlx5dv_devx_obj *devx_obj,
 				     uint32_t offset)
@@ -1174,7 +1346,7 @@ mlx5dv_dr_action_create_aso(struct mlx5dv_dr_domain *dmn,
 	struct mlx5dv_dr_action *action = NULL;
 
 	if (!dmn->info.supp_sw_steering ||
-	    dmn->info.caps.sw_format_ver != MLX5_HW_CONNECTX_6DX) {
+	    dmn->info.caps.sw_format_ver == MLX5_HW_CONNECTX_5) {
 		errno = EOPNOTSUPP;
 		return NULL;
 	}
@@ -1354,14 +1526,13 @@ out_err:
 	return errno;
 }
 
-#define ACTION_CACHE_LINE_SIZE 64
-
 static int
 dr_action_create_reformat_action(struct mlx5dv_dr_domain *dmn,
 				 size_t data_sz, void *data,
 				 struct mlx5dv_dr_action *action)
 {
 	struct mlx5dv_devx_obj *obj;
+	uint8_t *hw_actions;
 
 	switch (action->action_type) {
 	case DR_ACTION_TYP_L2_TO_TNL_L2:
@@ -1388,36 +1559,31 @@ dr_action_create_reformat_action(struct mlx5dv_dr_domain *dmn,
 	}
 	case DR_ACTION_TYP_TNL_L3_TO_L2:
 	{
-		uint8_t hw_actions[ACTION_CACHE_LINE_SIZE] = {};
 		int ret;
 
+		hw_actions = calloc(1, ACTION_CACHE_LINE_SIZE);
+		if (!hw_actions) {
+			errno = ENOMEM;
+			return errno;
+		}
+
 		ret = dr_ste_set_action_decap_l3_list(dmn->ste_ctx,
 						      data, data_sz,
 						      hw_actions,
 						      ACTION_CACHE_LINE_SIZE,
-						      &action->rewrite.num_of_actions);
+						      &action->rewrite.param.num_of_actions);
 		if (ret) {
 			dr_dbg(dmn, "Failed creating decap l3 action list\n");
-			return ret;
+			goto free_hw_actions;
 		}
 
-		action->rewrite.chunk = dr_icm_alloc_chunk(dmn->action_icm_pool,
-							   DR_CHUNK_SIZE_8);
-		if (!action->rewrite.chunk) {
-			dr_dbg(dmn, "Failed allocating modify header chunk\n");
-			return errno;
-		}
+		action->rewrite.param.data = hw_actions;
+		action->rewrite.dmn = dmn;
 
-		action->rewrite.data = (void *)hw_actions;
-		action->rewrite.index = (action->rewrite.chunk->icm_addr -
-					 dmn->info.caps.hdr_modify_icm_addr) /
-					 ACTION_CACHE_LINE_SIZE;
-
-		ret = dr_send_postsend_action(dmn, action);
+		ret = dr_ste_alloc_modify_hdr(action);
 		if (ret) {
-			dr_dbg(dmn, "Writing decap l3 actions to ICM failed\n");
-			dr_icm_free_chunk(action->rewrite.chunk);
-			return ret;
+			dr_dbg(dmn, "Failed prepare reformat data\n");
+			goto free_hw_actions;
 		}
 		return 0;
 	}
@@ -1426,6 +1592,9 @@ dr_action_create_reformat_action(struct mlx5dv_dr_domain *dmn,
 		errno = ENOTSUP;
 		return errno;
 	}
+free_hw_actions:
+	free(hw_actions);
+	return errno;
 }
 
 struct mlx5dv_dr_action *
@@ -1949,8 +2118,6 @@ static int dr_action_create_modify_action(struct mlx5dv_dr_domain *dmn,
 					  __be64 actions[],
 					  struct mlx5dv_dr_action *action)
 {
-	uint32_t dynamic_chunck_size;
-	struct dr_icm_chunk *chunk;
 	uint32_t num_hw_actions;
 	uint32_t num_sw_actions;
 	__be64 *hw_actions;
@@ -1978,31 +2145,22 @@ static int dr_action_create_modify_action(struct mlx5dv_dr_domain *dmn,
 	if (ret)
 		goto free_hw_actions;
 
-	dynamic_chunck_size = ilog32(num_hw_actions - 1);
+	action->rewrite.param.data = (uint8_t *)hw_actions;
+	action->rewrite.param.num_of_actions = num_hw_actions;
 
-	/* HW modify action index granularity is at least 64B */
-	dynamic_chunck_size = max_t(uint32_t, dynamic_chunck_size,
-				    DR_CHUNK_SIZE_8);
+	if (num_hw_actions == 1 &&
+	    (dmn->ste_ctx->actions_caps &
+	     DR_STE_CTX_ACTION_CAP_MODIFY_HDR_INLINE)) {
+		action->rewrite.single_action_opt = true;
+		return 0;
+	}
 
-	chunk = dr_icm_alloc_chunk(dmn->action_icm_pool, dynamic_chunck_size);
-	if (!chunk)
-		goto free_hw_actions;
-
-	action->rewrite.chunk = chunk;
-	action->rewrite.data = (uint8_t *)hw_actions;
-	action->rewrite.num_of_actions = num_hw_actions;
-	action->rewrite.index = (chunk->icm_addr -
-				 dmn->info.caps.hdr_modify_icm_addr) /
-				 ACTION_CACHE_LINE_SIZE;
-
-	ret = dr_send_postsend_action(dmn, action);
+	ret = dr_ste_alloc_modify_hdr(action);
 	if (ret)
-		goto free_chunk;
+		goto free_hw_actions;
 
 	return 0;
 
-free_chunk:
-	dr_icm_free_chunk(chunk);
 free_hw_actions:
 	free(hw_actions);
 	return errno;
@@ -2092,7 +2250,7 @@ struct mlx5dv_dr_action *
 mlx5dv_dr_action_create_flow_meter(struct mlx5dv_dr_flow_meter_attr *attr)
 {
 	struct mlx5dv_dr_domain *dmn = attr->next_table->dmn;
-	uint64_t rx_icm_addr, tx_icm_addr;
+	uint64_t rx_icm_addr = 0, tx_icm_addr = 0;
 	struct mlx5dv_devx_obj *devx_obj;
 	struct mlx5dv_dr_action *action;
 	int ret;
@@ -2402,7 +2560,7 @@ dr_action_create_sampler(struct mlx5dv_dr_domain *dmn,
 {
 	struct dr_devx_flow_sampler_attr sampler_attr = {};
 	struct dr_flow_sampler *sampler;
-	uint64_t icm_rx, icm_tx;
+	uint64_t icm_rx = 0, icm_tx = 0;
 	int ret;
 
 	sampler = calloc(1, sizeof(struct dr_flow_sampler));
@@ -2814,10 +2972,12 @@ int mlx5dv_dr_action_destroy(struct mlx5dv_dr_action *action)
 		atomic_fetch_sub(&action->reformat.dmn->refcount, 1);
 		break;
 	case DR_ACTION_TYP_TNL_L3_TO_L2:
-		if (action->reformat.is_root_level)
+		if (action->reformat.is_root_level) {
 			mlx5_destroy_flow_action(action->reformat.flow_action);
-		else
-			dr_icm_free_chunk(action->rewrite.chunk);
+		} else {
+			dr_ste_free_modify_hdr(action);
+			free(action->rewrite.param.data);
+		}
 		atomic_fetch_sub(&action->reformat.dmn->refcount, 1);
 		break;
 	case DR_ACTION_TYP_L2_TO_TNL_L2:
@@ -2832,8 +2992,10 @@ int mlx5dv_dr_action_destroy(struct mlx5dv_dr_action *action)
 		if (action->rewrite.is_root_level) {
 			mlx5_destroy_flow_action(action->rewrite.flow_action);
 		} else {
-			dr_icm_free_chunk(action->rewrite.chunk);
-			free(action->rewrite.data);
+			if (!action->rewrite.single_action_opt)
+				dr_ste_free_modify_hdr(action);
+
+			free(action->rewrite.param.data);
 		}
 		atomic_fetch_sub(&action->rewrite.dmn->refcount, 1);
 		break;
@@ -2855,6 +3017,11 @@ int mlx5dv_dr_action_destroy(struct mlx5dv_dr_action *action)
 		dr_action_remove_action_members(&action->dest_array.actions_list);
 		atomic_fetch_sub(&action->dest_array.dmn->refcount, 1);
 		break;
+	case DR_ACTION_TYP_ROOT_FT:
+		dr_devx_destroy_always_hit_ft(action->root_tbl.devx_tbl);
+		mlx5dv_destroy_steering_anchor(action->root_tbl.sa);
+		atomic_fetch_sub(&action->root_tbl.tbl->refcount, 1);
+		break;
 	default:
 		break;
 	}
