@@ -406,6 +406,7 @@ struct mlx5_context {
 	__be32                          dump_fill_mkey_be;
 	uint32_t			flags;
 	struct list_head		dyn_uar_bf_list;
+	struct list_head		dyn_uar_db_list;
 	struct list_head		dyn_uar_qp_shared_list;
 	struct list_head		dyn_uar_qp_dedicated_list;
 	uint16_t			qp_max_dedicated_uuars;
@@ -419,6 +420,8 @@ struct mlx5_context {
 	struct mlx5_dv_context_ops	*dv_ctx_ops;
 	struct mlx5dv_devx_obj		*crypto_login;
 	pthread_mutex_t			crypto_login_mutex;
+	uint64_t			max_dc_rd_atom;
+	uint64_t			max_dc_init_rd_atom;
 };
 
 struct mlx5_hugetlb_mem {
@@ -617,6 +620,7 @@ struct mlx5_bf {
 	uint8_t				dyn_alloc_uar : 1;
 	uint8_t				mmaped_entry : 1;
 	uint8_t				nc_mode : 1;
+	uint8_t				singleton : 1;
 	uint8_t				qp_dedicated : 1;
 	uint8_t				qp_shared : 1;
 	uint32_t			count;
