@@ -806,7 +806,6 @@ static void wr_atomic_cmp_swp(struct ibv_qp_ex *ibqp, uint32_t rkey,
 	wqe->wr.wr.atomic.swap = swap;
 	wqe->wr.wr.atomic.rkey = rkey;
 	wqe->iova = remote_addr;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -829,7 +828,6 @@ static void wr_atomic_fetch_add(struct ibv_qp_ex *ibqp, uint32_t rkey,
 	wqe->wr.wr.atomic.compare_add = add;
 	wqe->wr.wr.atomic.rkey = rkey;
 	wqe->iova = remote_addr;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -854,7 +852,6 @@ static void wr_bind_mw(struct ibv_qp_ex *ibqp, struct ibv_mw *ibmw,
 	wqe->wr.wr.mw.mw_rkey = ibmw->rkey;
 	wqe->wr.wr.mw.rkey = rkey;
 	wqe->wr.wr.mw.access = info->mw_access_flags;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -873,7 +870,6 @@ static void wr_local_inv(struct ibv_qp_ex *ibqp, uint32_t invalidate_rkey)
 	wqe->wr.opcode = IBV_WR_LOCAL_INV;
 	wqe->wr.send_flags = qp->vqp.qp_ex.wr_flags;
 	wqe->wr.ex.invalidate_rkey = invalidate_rkey;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -895,7 +891,6 @@ static void wr_rdma_read(struct ibv_qp_ex *ibqp, uint32_t rkey,
 	wqe->wr.wr.rdma.remote_addr = remote_addr;
 	wqe->wr.wr.rdma.rkey = rkey;
 	wqe->iova = remote_addr;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -917,7 +912,32 @@ static void wr_rdma_write(struct ibv_qp_ex *ibqp, uint32_t rkey,
 	wqe->wr.wr.rdma.remote_addr = remote_addr;
 	wqe->wr.wr.rdma.rkey = rkey;
 	wqe->iova = remote_addr;
-	wqe->ssn = qp->ssn++;
+
+	advance_qp_cur_index(qp);
+}
+
+static void wr_flush(struct ibv_qp_ex *ibqp, uint32_t rkey,
+		     uint64_t remote_addr, size_t length,
+		     uint8_t type, uint8_t level)
+{
+	struct rxe_qp *qp = container_of(ibqp, struct rxe_qp, vqp.qp_ex);
+	struct rxe_send_wqe *wqe = addr_from_index(qp->sq.queue, qp->cur_index);
+
+	if (check_qp_queue_full(qp))
+		return;
+
+	memset(wqe, 0, sizeof(*wqe));
+
+	wqe->wr.wr_id = qp->vqp.qp_ex.wr_id;
+	wqe->wr.opcode = IBV_WR_FLUSH;
+	wqe->wr.send_flags = qp->vqp.qp_ex.wr_flags;
+	wqe->wr.wr.flush.remote_addr = remote_addr;
+	wqe->wr.wr.flush.rkey = rkey;
+	wqe->wr.wr.flush.type = type;
+	wqe->wr.wr.flush.level = level;
+	wqe->dma.length = length;
+	wqe->dma.resid = length;
+	wqe->iova = remote_addr;
 
 	advance_qp_cur_index(qp);
 }
@@ -943,7 +963,6 @@ static void wr_atomic_write(struct ibv_qp_ex *ibqp, uint32_t rkey,
 	wqe->dma.resid = 8;
 
 	wqe->iova = remote_addr;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -966,7 +985,6 @@ static void wr_rdma_write_imm(struct ibv_qp_ex *ibqp, uint32_t rkey,
 	wqe->wr.wr.rdma.rkey = rkey;
 	wqe->wr.ex.imm_data = imm_data;
 	wqe->iova = remote_addr;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -984,7 +1002,6 @@ static void wr_send(struct ibv_qp_ex *ibqp)
 	wqe->wr.wr_id = qp->vqp.qp_ex.wr_id;
 	wqe->wr.opcode = IBV_WR_SEND;
 	wqe->wr.send_flags = qp->vqp.qp_ex.wr_flags;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -1003,7 +1020,6 @@ static void wr_send_imm(struct ibv_qp_ex *ibqp, __be32 imm_data)
 	wqe->wr.opcode = IBV_WR_SEND_WITH_IMM;
 	wqe->wr.send_flags = qp->vqp.qp_ex.wr_flags;
 	wqe->wr.ex.imm_data = imm_data;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -1022,7 +1038,6 @@ static void wr_send_inv(struct ibv_qp_ex *ibqp, uint32_t invalidate_rkey)
 	wqe->wr.opcode = IBV_WR_SEND_WITH_INV;
 	wqe->wr.send_flags = qp->vqp.qp_ex.wr_flags;
 	wqe->wr.ex.invalidate_rkey = invalidate_rkey;
-	wqe->ssn = qp->ssn++;
 
 	advance_qp_cur_index(qp);
 }
@@ -1266,7 +1281,7 @@ enum {
 		IBV_QP_EX_WITH_RDMA_READ | IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP |
 		IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD | IBV_QP_EX_WITH_LOCAL_INV |
 		IBV_QP_EX_WITH_BIND_MW | IBV_QP_EX_WITH_SEND_WITH_INV |
-		IBV_QP_EX_WITH_ATOMIC_WRITE,
+		IBV_QP_EX_WITH_FLUSH | IBV_QP_EX_WITH_ATOMIC_WRITE,
 
 	RXE_SUP_UC_QP_SEND_OPS_FLAGS =
 		IBV_QP_EX_WITH_RDMA_WRITE | IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM |
@@ -1334,6 +1349,9 @@ static void set_qp_send_ops(struct rxe_qp *qp, uint64_t flags)
 	if (flags & IBV_QP_EX_WITH_RDMA_WRITE)
 		qp->vqp.qp_ex.wr_rdma_write = wr_rdma_write;
 
+	if (flags & IBV_QP_EX_WITH_FLUSH)
+		qp->vqp.qp_ex.wr_flush = wr_flush;
+
 	if (flags & IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM)
 		qp->vqp.qp_ex.wr_rdma_write_imm = wr_rdma_write_imm;
 
@@ -1569,7 +1587,6 @@ static int init_send_wqe(struct rxe_qp *qp, struct rxe_wq *sq,
 	wqe->dma.cur_sge	= 0;
 	wqe->dma.sge_offset	= 0;
 	wqe->state		= 0;
-	wqe->ssn		= qp->ssn++;
 
 	return 0;
 }
