@@ -16,21 +16,21 @@ import glob
 import time
 import os
 
+from pyverbs.qp import QPCap, QPInitAttr, QPInitAttrEx, QPAttr, QPEx, QP
 from pyverbs.pyverbs_error import PyverbsError, PyverbsRDMAError
-from pyverbs.addr import AHAttr, AH, GlobalRoute
+from tests.mlx5_base import Mlx5DcResources, Mlx5DcStreamsRes
 from tests.base import XRCResources, DCT_KEY, MLNX_VENDOR_ID
-from tests.efa_base import SRDResources
+from pyverbs.addr import AHAttr, AH, GlobalRoute
 from pyverbs.providers.efa.efadv import EfaCQ
 from pyverbs.wr import SGE, SendWR, RecvWR
-from pyverbs.qp import QPCap, QPInitAttr, QPInitAttrEx, QPAttr, QPEx, QP
-from tests.mlx5_base import Mlx5DcResources, Mlx5DcStreamsRes
 from pyverbs.base import PyverbsRDMAErrno
+from tests.efa_base import SRDResources
 from pyverbs.cq import PollCqAttr, CQEX
 from pyverbs.mr import MW, MWBindInfo
+from pyverbs.mem_alloc import madvise
 import pyverbs.device as d
 import pyverbs.enums as e
 from pyverbs.mr import MR
-import pyverbs.providers.mlx5.mlx5_enums as me
 
 
 MAX_MR_SIZE = 4194304
@@ -77,6 +77,7 @@ class PacketConsts:
     TTL_HOP_LIMIT = 64
     IHL = 5
     # Hardcoded values for flow matchers
+    ETHER_TYPE_ETH = 0x6558
     ETHER_TYPE_IPV4 = 0x800
     MAC_MASK = "ff:ff:ff:ff:ff:ff"
     ETHER_TYPE_IPV6 = 0x86DD
@@ -99,6 +100,18 @@ class PacketConsts:
     VLAN_PRIO = 5
     VLAN_CFI = 1
     VLAN_ID = 0xc0c
+    GRE_VER = 1
+    GRE_FLAGS = 2
+    GRE_KEY = 0x12345678
+    GENEVE_VNI = 2
+    GENEVE_OAM = 0
+    GENEVE_PORT = 6081
+    BTH_HEADER_SIZE = 16
+    BTH_OPCODE = 0x81
+    BTH_DST_QP = 0xd2
+    BTH_PARTITION_KEY = 0xffff
+    BTH_BECN = 1
+    ROCE_PORT = 4791
 
 
 def get_mr_length():
@@ -363,7 +376,7 @@ def wc_status_to_str(status):
         return 'Unknown WC status ({s})'.format(s=status)
 
 
-def create_custom_mr(agr_obj, additional_access_flags=0, size=None):
+def create_custom_mr(agr_obj, additional_access_flags=0, size=None, user_addr=None):
     """
     Creates a memory region using the aggregation object's PD.
     If size is None, the agr_obj's message size is used to set the MR's size.
@@ -371,11 +384,12 @@ def create_custom_mr(agr_obj, additional_access_flags=0, size=None):
     :param agr_obj: The aggregation object that creates the MR
     :param additional_access_flags: Addition access flags to set in the MR
     :param size: MR's length. If None, agr_obj.msg_size is used.
+    :param user_addr: The MR's buffer address. If None, the buffer will be allocated by pyverbs.
     """
     mr_length = size if size else agr_obj.msg_size
     try:
         return MR(agr_obj.pd, mr_length,
-                  e.IBV_ACCESS_LOCAL_WRITE | additional_access_flags)
+                  e.IBV_ACCESS_LOCAL_WRITE | additional_access_flags, address=user_addr)
     except PyverbsRDMAError as ex:
         if ex.error_code == errno.EOPNOTSUPP:
                 raise unittest.SkipTest(f'Create custom mr with additional access flags {additional_access_flags} is not supported')
@@ -391,7 +405,7 @@ def get_send_elements(agr_obj, is_server, opcode=e.IBV_WR_SEND):
     :param is_server: Indicates whether this is server or client side
     :return: send wr and its SGE
     """
-    if opcode == e.IBV_QP_EX_WITH_ATOMIC_WRITE:
+    if opcode == e.IBV_WR_ATOMIC_WRITE:
         atomic_wr = agr_obj.msg_size * (b's' if is_server else b'c')
         return None, atomic_wr
 
@@ -402,8 +416,8 @@ def get_send_elements(agr_obj, is_server, opcode=e.IBV_WR_SEND):
     agr_obj.mem_write(msg, agr_obj.msg_size + offset)
     sge = SGE(agr_obj.mr.buf + offset, agr_obj.msg_size, agr_obj.mr_lkey)
     send_wr = SendWR(opcode=opcode, num_sge=1, sg=[sge])
-    if opcode in [e.IBV_WR_RDMA_WRITE, e.IBV_WR_RDMA_READ]:
-        send_wr.set_wr_rdma(int(agr_obj.rkey), int(agr_obj.remote_addr))
+    if opcode in [e.IBV_WR_RDMA_WRITE, e.IBV_WR_RDMA_WRITE_WITH_IMM, e.IBV_WR_RDMA_READ]:
+        send_wr.set_wr_rdma(int(agr_obj.rkey), int(agr_obj.raddr))
     return send_wr, sge
 
 def get_recv_wr(agr_obj):
@@ -454,31 +468,37 @@ def xrc_post_send(agr_obj, qp_num, send_object, send_op=None):
         post_send(agr_obj, send_object)
 
 
-def post_send_ex(agr_obj, send_object, send_op=None, qp_idx=0, ah=None):
+def post_send_ex(agr_obj, send_object, send_op=None, qp_idx=0, ah=None, **kwargs):
     qp = agr_obj.qps[qp_idx]
     qp_type = qp.qp_type
     qp.wr_start()
     qp.wr_id = 0x123
     qp.wr_flags = e.IBV_SEND_SIGNALED
-    if send_op == e.IBV_QP_EX_WITH_SEND:
+    if send_op == e.IBV_WR_SEND:
         qp.wr_send()
-    elif send_op == e.IBV_QP_EX_WITH_RDMA_WRITE:
+    elif send_op == e.IBV_WR_RDMA_WRITE:
         qp.wr_rdma_write(agr_obj.rkey, agr_obj.raddr)
-    elif send_op == e.IBV_QP_EX_WITH_SEND_WITH_IMM:
+    elif send_op == e.IBV_WR_SEND_WITH_IMM:
         qp.wr_send_imm(IMM_DATA)
-    elif send_op == e.IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM:
+    elif send_op == e.IBV_WR_RDMA_WRITE_WITH_IMM:
         qp.wr_rdma_write_imm(agr_obj.rkey, agr_obj.raddr, IMM_DATA)
-    elif send_op == e.IBV_QP_EX_WITH_ATOMIC_WRITE:
+    elif send_op == e.IBV_WR_ATOMIC_WRITE:
         qp.wr_atomic_write(agr_obj.rkey, agr_obj.raddr, send_object)
-    elif send_op == e.IBV_QP_EX_WITH_RDMA_READ:
+    elif send_op == e.IBV_WR_FLUSH:
+        qp.wr_flush(agr_obj.rkey, agr_obj.raddr, agr_obj.msg_size,
+                    agr_obj.ptype, agr_obj.level)
+    elif send_op == e.IBV_WR_RDMA_READ:
         qp.wr_rdma_read(agr_obj.rkey, agr_obj.raddr)
-    elif send_op == e.IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP:
+    elif send_op == e.IBV_WR_ATOMIC_CMP_AND_SWP:
+        cmp_add = kwargs.get('cmp_add')
+        swp = kwargs.get('swap')
         qp.wr_atomic_cmp_swp(agr_obj.rkey, agr_obj.raddr,
-                             int8b_from_int(2), int8b_from_int(0))
-    elif send_op == e.IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD:
+                             int8b_from_int(cmp_add), int8b_from_int(swp))
+    elif send_op == e.IBV_WR_ATOMIC_FETCH_AND_ADD:
+        cmp_add = kwargs.get('cmp_add')
         qp.wr_atomic_fetch_add(agr_obj.rkey, agr_obj.raddr,
-                               int8b_from_int(2))
-    elif send_op == e.IBV_QP_EX_WITH_BIND_MW:
+                               int8b_from_int(cmp_add))
+    elif send_op == e.IBV_WR_BIND_MW:
         bind_info = MWBindInfo(agr_obj.mr, agr_obj.mr.buf, agr_obj.mr.rkey,
                                e.IBV_ACCESS_REMOTE_WRITE)
         mw = MW(agr_obj.pd, mw_type=e.IBV_MW_TYPE_2)
@@ -499,7 +519,8 @@ def post_send_ex(agr_obj, send_object, send_op=None, qp_idx=0, ah=None):
                                      stream_id)
         else:
             qp.wr_set_dc_addr(ah, agr_obj.remote_dct_num, DCT_KEY)
-    if send_op != e.IBV_QP_EX_WITH_ATOMIC_WRITE:
+    if send_op != e.IBV_WR_ATOMIC_WRITE and \
+            send_op != e.IBV_WR_FLUSH:
         qp.wr_set_sge(send_object)
     qp.wr_complete()
 
@@ -679,16 +700,17 @@ def validate(received_str, is_server, msg_size):
                 format(exp=expected_str, rcv=received_str))
 
 
-def send(agr_obj, send_object, send_op=None, new_send=False, qp_idx=0, ah=None, is_imm=False):
+def send(agr_obj, send_object, send_op=None, new_send=False, qp_idx=0, ah=None, is_imm=False,
+         **kwargs):
     if isinstance(agr_obj, XRCResources):
         agr_obj.qps = agr_obj.sqp_lst
     if new_send:
-        return post_send_ex(agr_obj, send_object, send_op, qp_idx, ah)
+        return post_send_ex(agr_obj, send_object, send_op, qp_idx, ah, **kwargs)
     return post_send(agr_obj, send_object, qp_idx, ah, is_imm)
 
 
-def traffic(client, server, iters, gid_idx, port, is_cq_ex=False, send_op=None,
-            new_send=False, is_imm=False):
+def traffic(client, server, iters, gid_idx, port, is_cq_ex=False, send_op=e.IBV_WR_SEND,
+            new_send=False, force_page_faults=False):
     """
     Runs basic traffic between two sides
     :param client: client side, clients base class is BaseTraffic
@@ -699,7 +721,8 @@ def traffic(client, server, iters, gid_idx, port, is_cq_ex=False, send_op=None,
     :param is_cq_ex: If True, use poll_cq_ex() rather than poll_cq()
     :param send_op: The send_wr opcode.
     :param new_send: If True use new post send API.
-    :param is_imm: If True, send with imm_data, relevant for old post send API.
+    :param force_page_faults: If True, use madvise to hint that we don't need the MR's buffer to
+                              force page faults (useful for ODP testing).
     :return:
     """
     if is_datagram_qp(client):
@@ -709,12 +732,11 @@ def traffic(client, server, iters, gid_idx, port, is_cq_ex=False, send_op=None,
         ah_client = None
         ah_server = None
     poll = poll_cq_ex if is_cq_ex else poll_cq
-    if send_op == e.IBV_QP_EX_WITH_SEND_WITH_IMM or \
-       send_op == e.IBV_QP_EX_WITH_RDMA_WRITE_WITH_IMM or \
-       is_imm:
+
+    imm_data = None
+    if send_op in [e.IBV_WR_SEND_WITH_IMM, e.IBV_WR_RDMA_WRITE_WITH_IMM]:
         imm_data = IMM_DATA
-    else:
-        imm_data = None
+
     s_recv_wr = get_recv_wr(server)
     c_recv_wr = get_recv_wr(client)
     for qp_idx in range(server.qp_count):
@@ -724,37 +746,34 @@ def traffic(client, server, iters, gid_idx, port, is_cq_ex=False, send_op=None,
     read_offset = GRH_SIZE if client.qp.qp_type == e.IBV_QPT_UD else 0
     for _ in range(iters):
         for qp_idx in range(server.qp_count):
-            if imm_data:
-                c_send_wr, c_sg = get_send_elements(client, False, opcode=e.IBV_WR_SEND_WITH_IMM)
-            else:
-                c_send_wr, c_sg = get_send_elements(client, False)
+            if force_page_faults:
+                madvise(client.mr.buf, client.msg_size)
+                madvise(server.mr.buf, server.msg_size)
+            c_send_wr, c_sg = get_send_elements(client, False, send_op)
             if client.use_mr_prefetch:
                 flags = e._IBV_ADVISE_MR_FLAG_FLUSH
                 if client.use_mr_prefetch == 'async':
                     flags = 0
                 prefetch_mrs(client, [c_sg], advice=client.prefetch_advice,
                              flags=flags)
-            c_send_object = c_sg if send_op else c_send_wr
+            c_send_object = c_sg if new_send else c_send_wr
             send(client, c_send_object, send_op, new_send, qp_idx,
-                 ah_client, is_imm=is_imm)
+                 ah_client, is_imm=(imm_data != None))
             poll(client.cq)
             poll(server.cq, data=imm_data)
             post_recv(server, s_recv_wr, qp_idx=qp_idx)
             msg_received = server.mr.read(server.msg_size, read_offset)
             validate(msg_received, True, server.msg_size)
-            if imm_data:
-                s_send_wr, s_sg = get_send_elements(server, True, opcode=e.IBV_WR_SEND_WITH_IMM)
-            else:
-                s_send_wr, s_sg = get_send_elements(server, True)
+            s_send_wr, s_sg = get_send_elements(server, True, send_op)
             if server.use_mr_prefetch:
                 flags = e._IBV_ADVISE_MR_FLAG_FLUSH
                 if server.use_mr_prefetch == 'async':
                     flags = 0
                 prefetch_mrs(server, [s_sg], advice=server.prefetch_advice,
                              flags=flags)
-            s_send_object = s_sg if send_op else s_send_wr
+            s_send_object = s_sg if new_send else s_send_wr
             send(server, s_send_object, send_op, new_send, qp_idx,
-                 ah_server, is_imm=is_imm)
+                 ah_server, is_imm=(imm_data != None))
             poll(server.cq)
             poll(client.cq, data=imm_data)
             post_recv(client, c_recv_wr, qp_idx=qp_idx)
@@ -762,35 +781,91 @@ def traffic(client, server, iters, gid_idx, port, is_cq_ex=False, send_op=None,
             validate(msg_received, False, client.msg_size)
 
 
-def gen_outer_headers(msg_size):
+def gen_ethernet_header(dst_mac=PacketConsts.DST_MAC, src_mac=PacketConsts.SRC_MAC,
+                        ether_type=PacketConsts.ETHER_TYPE_IPV4):
     """
-    Generates outer headers for encapsulation with VXLAN: Ethernet, IPv4, UDP
-    and VXLAN using the values from the PacketConst class.
-    :param msg_size: The size of the inner message
-    :return: Outer headers
+    Generates Ethernet header using the values from the PacketConst class by default.
+    :param dst_mac: Destination mac address
+    :param src_mac: Source mac address
+    :param ether_type: Ether type of next header
+    :return: Ethernet header
     """
-    # Ethernet Header
-    outer = struct.pack('!6s6s',
-                        bytes.fromhex(PacketConsts.DST_MAC.replace(':', '')),
-                        bytes.fromhex(PacketConsts.SRC_MAC.replace(':', '')))
-    outer += PacketConsts.ETHER_TYPE_IPV4.to_bytes(2, 'big')
-    # IPv4 Header
-    ip_total_len = msg_size + PacketConsts.UDP_HEADER_SIZE + \
-                   PacketConsts.IPV4_HEADER_SIZE + \
-                   PacketConsts.VXLAN_HEADER_SIZE
-    outer += struct.pack('!2B3H2BH4s4s', (PacketConsts.IP_V4 << 4) +
-                         PacketConsts.IHL, 0, ip_total_len, 0,
-                         PacketConsts.IP_V4_FLAGS << 13,
-                         PacketConsts.TTL_HOP_LIMIT, socket.IPPROTO_UDP, 0,
-                         socket.inet_aton(PacketConsts.SRC_IP),
-                         socket.inet_aton(PacketConsts.DST_IP))
-    # UDP Header
-    outer += struct.pack('!4H', PacketConsts.SRC_PORT, PacketConsts.VXLAN_PORT,
-                         msg_size + PacketConsts.UDP_HEADER_SIZE + 8, 0)
-    # VXLAN Header
-    outer += struct.pack('!II', PacketConsts.VXLAN_FLAGS << 24,
-                         PacketConsts.VXLAN_VNI << 8)
-    return outer
+    header = struct.pack('!6s6s',
+                        bytes.fromhex(dst_mac.replace(':', '')),
+                        bytes.fromhex(src_mac.replace(':', '')))
+    header += ether_type.to_bytes(2, 'big')
+    return header
+
+
+def gen_ipv4_header(packet_len, next_proto=socket.IPPROTO_UDP, src_ip=PacketConsts.SRC_IP,
+                    dst_ip=PacketConsts.DST_IP):
+    """
+    Generates IPv4 header using the values from the PacketConst class by default.
+    :param packet_len: Length of all fields following the IP header
+    :param next_proto: protocol type of next header
+    :param src_ip: Source mac address
+    :param dst_ip: Destination mac address
+    :return: IPv4 header
+    """
+    ip_total_len = packet_len + PacketConsts.IPV4_HEADER_SIZE
+    return struct.pack('!2B3H2BH4s4s', (PacketConsts.IP_V4 << 4) +
+                       PacketConsts.IHL, 0, ip_total_len, 0,
+                       PacketConsts.IP_V4_FLAGS << 13,
+                       PacketConsts.TTL_HOP_LIMIT, next_proto, 0,
+                       socket.inet_aton(src_ip),
+                       socket.inet_aton(dst_ip))
+
+
+def gen_udp_header(packet_len, src_port=PacketConsts.SRC_PORT, dst_port=PacketConsts.DST_PORT):
+    """
+    Generates UDP header using the values from the PacketConst class by default.
+    :param packet_len: Length of all fields following the UDP header
+    :param src_port: Source port
+    :param dst_port: Destination port
+    :return: UDP header
+    """
+    udp_total_len = packet_len + PacketConsts.UDP_HEADER_SIZE
+    return struct.pack('!4H', src_port, dst_port, udp_total_len, 0)
+
+
+def gen_gre_header(ether_type=PacketConsts.ETHER_TYPE_IPV4):
+    """
+    Generates GRE header using the values from the PacketConst class by default.
+    :param ether_type: Ether type of tunneled next header
+    :return: GRE header
+    """
+    return struct.pack('!2BHI', PacketConsts.GRE_FLAGS << 4, PacketConsts.GRE_VER,
+                       ether_type, PacketConsts.GRE_KEY)
+
+
+def gen_vxlan_header():
+    """
+    Generates VXLAN header using the values from the PacketConst class by default.
+    :return: VXLAN header
+    """
+    return struct.pack('!II', PacketConsts.VXLAN_FLAGS << 24, PacketConsts.VXLAN_VNI << 8)
+
+
+def gen_geneve_header(vni=PacketConsts.GENEVE_VNI, oam=PacketConsts.GENEVE_OAM,
+                      proto=PacketConsts.ETHER_TYPE_ETH):
+    """
+    Generates Geneve header using the values from the PacketConst class by default.
+    :param vni: geneve vni
+    :param oam: geneve oam
+    :param proto: Ether type of next header inside the tunnel
+    :return: Geneve header
+    """
+    return struct.pack('!BBHL', (0 << 6) + 0, (oam << 7) + (0 << 6) + 0, proto, (vni << 8) + 0)
+
+def gen_bth_header(opcode=PacketConsts.BTH_OPCODE, dst_qp=PacketConsts.BTH_DST_QP):
+    """
+    Generates ROCE BTH header using the values from the PacketConst class by default.
+    :param opcode: BTH opcode
+    :param dst_qp: BTH dst QP
+    :return: ROCE BTH header
+    """
+    return struct.pack('!2BH2BH2L', opcode, 0, PacketConsts.BTH_PARTITION_KEY,
+                       PacketConsts.BTH_BECN << 6, dst_qp >> 16, dst_qp & 0xffff, 0, 0)
 
 
 def gen_packet(msg_size, l3=PacketConsts.IP_V4, l4=PacketConsts.UDP_PROTO, with_vlan=False, **kwargs):
@@ -863,7 +938,8 @@ def gen_packet(msg_size, l3=PacketConsts.IP_V4, l4=PacketConsts.UDP_PROTO, with_
 
 
 def get_send_elements_raw_qp(agr_obj, l3=PacketConsts.IP_V4,
-                             l4=PacketConsts.UDP_PROTO, with_vlan=False, **packet_args):
+                             l4=PacketConsts.UDP_PROTO, with_vlan=False,
+                             packet_to_send=None, **packet_args):
     """
     Creates a single SGE and a single Send WR for agr_obj's RAW QP type. The
     content of the message is Eth | Ipv4 | UDP packet.
@@ -871,11 +947,14 @@ def get_send_elements_raw_qp(agr_obj, l3=PacketConsts.IP_V4,
     :param l3: Packet layer 3 type: 4 for IPv4 or 6 for IPv6
     :param l4: Packet layer 4 type: 'tcp' or 'udp'
     :param with_vlan: if True add VLAN header to the packet
+    :param packet_to_send: If passed, the other packet related parameters would
+                           be ignored, and this will be the packet to send.
     :param packet_args: Pass packet_args to gen_packets method.
     :return: send wr, its SGE, and message
     """
     mr = agr_obj.mr
-    msg = gen_packet(agr_obj.msg_size, l3, l4, with_vlan, **packet_args)
+    msg = packet_to_send if packet_to_send is not None else \
+        gen_packet(agr_obj.msg_size, l3, l4, with_vlan, **packet_args)
     mr.write(msg, agr_obj.msg_size)
     sge = SGE(mr.buf, agr_obj.msg_size, mr.lkey)
     send_wr = SendWR(opcode=e.IBV_WR_SEND, num_sge=1, sg=[sge])
@@ -915,7 +994,7 @@ def sampler_traffic(client, server, iters, l3=PacketConsts.IP_V4, l4=PacketConst
 
 def raw_traffic(client, server, iters, l3=PacketConsts.IP_V4,
                 l4=PacketConsts.UDP_PROTO, with_vlan=False, expected_packet=None,
-                skip_idxs=None):
+                skip_idxs=None, packet_to_send=None):
     """
     Runs raw ethernet traffic between two sides
     :param client: client side, clients base class is BaseTraffic
@@ -927,7 +1006,8 @@ def raw_traffic(client, server, iters, l3=PacketConsts.IP_V4,
     :param expected_packet: Expected packet for validation (when different from
                             the originally sent).
     :param skip_idxs: indexes to skip during packet validation
-    :return:
+    :param packet_to_send: If passed, the other packet related parameters would
+                           be ignored, and this will be the packet to send.
     """
     skip_idxs = [] if skip_idxs is None else skip_idxs
     s_recv_wr = get_recv_wr(server)
@@ -940,7 +1020,8 @@ def raw_traffic(client, server, iters, l3=PacketConsts.IP_V4,
     poll = poll_cq_ex if isinstance(client.cq, CQEX) else poll_cq
     for _ in range(iters):
         for qp_idx in range(server.qp_count):
-            c_send_wr, c_sg, msg = get_send_elements_raw_qp(client, l3, l4, with_vlan)
+            c_send_wr, c_sg, msg = get_send_elements_raw_qp(client, l3, l4, with_vlan,
+                                                            packet_to_send=packet_to_send)
             send(client, c_send_wr, e.IBV_WR_SEND, False, qp_idx)
             poll(client.cq)
             poll(server.cq)
@@ -996,8 +1077,41 @@ def raw_rss_traffic(client, server, iters, l3=PacketConsts.IP_V4,
             post_recv(server, s_recv_wr, qp_idx=qp_idx, num_wqes=num_packets)
 
 
+def flush_traffic(client, server, iters, gid_idx, port, new_send=False,
+                  send_op=None):
+    """
+    Runs basic RDMA FLUSH traffic that client requests a FLUSH to server.
+    Simply, run RDMA WRITE and then follow up by a RDMA FLUSH.
+    No receive WQEs are posted.
+    :param client: client side, clients base class is BaseTraffic
+    :param server: server side, servers base class is BaseTraffic
+    :param iters: number of traffic iterations
+    :param gid_idx: local gid index
+    :param port: IB port
+    :param new_send: If True use new post send API.
+    :param send_op: The send_wr opcode.
+    :return:
+    """
+    rdma_traffic(client, server, iters, gid_idx, port, new_send, e.IBV_WR_RDMA_WRITE)
+    for i in range(iters):
+        if client.level == e.IBV_FLUSH_MR:
+            client.msg_size = 0 if i == 0 else random.randint(0, 12345678)
+        send(client, None, send_op, new_send)
+        wcs = _poll_cq(client.cq)
+        if (wcs[0].status != e.IBV_WC_SUCCESS):
+            break
+    return wcs
+
+
+def prepare_validate_data(client=None, server=None):
+    if server:
+        server.mem_write('s' * server.msg_size, server.msg_size)
+    if client:
+        client.mem_write('c' * client.msg_size, client.msg_size)
+
+
 def rdma_traffic(client, server, iters, gid_idx, port, new_send=False,
-                 send_op=None):
+                 send_op=None, force_page_faults=False):
     """
     Runs basic RDMA traffic between two sides. No receive WQEs are posted. For
     RDMA send with immediate, use traffic().
@@ -1008,6 +1122,8 @@ def rdma_traffic(client, server, iters, gid_idx, port, new_send=False,
     :param port: IB port
     :param new_send: If True use new post send API.
     :param send_op: The send_wr opcode.
+    :param force_page_faults: If True, use madvise to hint that we don't need the MR's buffer to
+                              force page faults (useful for ODP testing).
     :return:
     """
     # Using the new post send API, we need the SGE, not the SendWR
@@ -1019,11 +1135,14 @@ def rdma_traffic(client, server, iters, gid_idx, port, new_send=False,
         ah_client = None
         ah_server = None
     send_element_idx = 1 if new_send else 0
-    same_side_check =  send_op in [e.IBV_QP_EX_WITH_RDMA_READ,
-                                   e.IBV_QP_EX_WITH_ATOMIC_CMP_AND_SWP,
-                                   e.IBV_QP_EX_WITH_ATOMIC_FETCH_AND_ADD,
-                                   e.IBV_WR_RDMA_READ]
+    same_side_check =  send_op in [e.IBV_WR_RDMA_READ,
+                                   e.IBV_WR_ATOMIC_CMP_AND_SWP,
+                                   e.IBV_WR_ATOMIC_FETCH_AND_ADD]
     for _ in range(iters):
+        if force_page_faults:
+            madvise(client.mr.buf, client.msg_size)
+            madvise(server.mr.buf, server.msg_size)
+        prepare_validate_data(client=client, server=server)
         c_send_wr = get_send_elements(client, False, send_op)[send_element_idx]
         send(client, c_send_wr, send_op, new_send, ah=ah_client)
         poll_cq(client.cq)
@@ -1034,8 +1153,7 @@ def rdma_traffic(client, server, iters, gid_idx, port, new_send=False,
         validate(msg_received, False if same_side_check else True,
                  server.msg_size)
         s_send_wr = get_send_elements(server, True, send_op)[send_element_idx]
-        if same_side_check:
-            client.mem_write('c' * client.msg_size, client.msg_size)
+        prepare_validate_data(client=client, server=server)
         send(server, s_send_wr, send_op, new_send, ah=ah_server)
         poll_cq(server.cq)
         if same_side_check:
@@ -1044,12 +1162,11 @@ def rdma_traffic(client, server, iters, gid_idx, port, new_send=False,
             msg_received = client.mem_read(server.msg_size)
         validate(msg_received, True if same_side_check else False,
                  client.msg_size)
-        if same_side_check:
-            server.mem_write('s' * server.msg_size, server.msg_size)
 
 
 def atomic_traffic(client, server, iters, gid_idx, port, new_send=False,
-                   send_op=None, receiver_val=1, sender_val=2):
+                   send_op=None, receiver_val=1, sender_val=2, swap=0,
+                   client_wr=1, server_wr=1, **kwargs):
     """
     Runs atomic traffic between two sides.
     :param client: Client side, clients base class is BaseTraffic
@@ -1061,31 +1178,49 @@ def atomic_traffic(client, server, iters, gid_idx, port, new_send=False,
     :param send_op: The send_wr opcode.
     :param receiver_val: The requested value on the reciver MR.
     :param sender_val: The requested value on the sender SendWR.
+    :param client_wr: Number of WR the client will post before polling all of them
+    :param server_wr: Number of WR the server will post before polling all of them
+    :param kwargs: General arguments (shared with other traffic functions).
     """
     send_element_idx = 1 if new_send else 0
+    if is_datagram_qp(client):
+        ah_client = get_global_ah(client, gid_idx, port)
+        ah_server = get_global_ah(server, gid_idx, port)
+    else:
+        ah_client = None
+        ah_server = None
+
     for _ in range(iters):
         client.mr.write(int.to_bytes(sender_val, 1, byteorder='big') * 8, 8)
         server.mr.write(int.to_bytes(receiver_val, 1, byteorder='big') * 8, 8)
-        c_send_wr = get_atomic_send_elements(client, send_op,
-                                             cmp_add=sender_val,
-                                             swap=0)[send_element_idx]
-        if isinstance(server, XRCResources):
-            c_send_wr.set_qp_type_xrc(server.srq.get_srq_num())
-        send(client, c_send_wr, send_op, new_send)
-        poll_cq(client.cq)
-        validate_atomic(send_op, server, client, receiver_val=receiver_val,
-                        send_cmp_add=sender_val, send_swp=0)
+        for _ in range(client_wr):
+            c_send_wr = get_atomic_send_elements(client,
+                                                 send_op,
+                                                 cmp_add=sender_val,
+                                                 swap=swap)[send_element_idx]
+            if isinstance(server, XRCResources):
+                c_send_wr.set_qp_type_xrc(server.srq.get_srq_num())
+            send(client, c_send_wr, send_op, new_send, ah=ah_client,
+                 cmp_add=sender_val, swap=swap)
+        poll_cq(client.cq, count=client_wr)
+        validate_atomic(send_op, server, client,
+                        receiver_val=receiver_val + sender_val * (client_wr - 1),
+                        send_cmp_add=sender_val, send_swp=swap)
         server.mr.write(int.to_bytes(sender_val, 1, byteorder='big') * 8, 8)
         client.mr.write(int.to_bytes(receiver_val, 1, byteorder='big') * 8, 8)
-        s_send_wr = get_atomic_send_elements(server, send_op,
-                                             cmp_add=sender_val,
-                                             swap=0)[send_element_idx]
-        if isinstance(client, XRCResources):
-            s_send_wr.set_qp_type_xrc(client.srq.get_srq_num())
-        send(server, s_send_wr, send_op, new_send)
-        poll_cq(server.cq)
-        validate_atomic(send_op, client, server, receiver_val=receiver_val,
-                        send_cmp_add=sender_val, send_swp=0)
+        for _ in range(server_wr):
+            s_send_wr = get_atomic_send_elements(server,
+                                                 send_op,
+                                                 cmp_add=sender_val,
+                                                 swap=swap)[send_element_idx]
+            if isinstance(client, XRCResources):
+                s_send_wr.set_qp_type_xrc(client.srq.get_srq_num())
+            send(server, s_send_wr, send_op, new_send, ah=ah_server,
+                 cmp_add=sender_val, swap=swap)
+        poll_cq(server.cq, count=server_wr)
+        validate_atomic(send_op, client, server,
+                        receiver_val=receiver_val + sender_val * (server_wr - 1),
+                        send_cmp_add=sender_val, send_swp=swap)
 
 
 def validate_atomic(opcode, recv_player, send_player, receiver_val,
@@ -1152,7 +1287,7 @@ def get_atomic_send_elements(agr_obj, opcode, cmp_add=0, swap=0):
     return send_wr, sge
 
 
-def xrc_traffic(client, server, is_cq_ex=False, send_op=None):
+def xrc_traffic(client, server, is_cq_ex=False, send_op=None, force_page_faults=False):
     """
     Runs basic xrc traffic, this function assumes that number of QPs, which
     server and client have are equal, server.send_qp[i] is connected to
@@ -1165,6 +1300,8 @@ def xrc_traffic(client, server, is_cq_ex=False, send_op=None):
     of XRCResources class
     :param is_cq_ex: If True, use poll_cq_ex() rather than poll_cq()
     :param send_op: If not None, new post send API is assumed.
+    :param force_page_faults: If True, use madvise to hint that we don't need the MR's buffer to
+                              force page faults (useful for ODP testing).
     :return: None
     """
     poll = poll_cq_ex if is_cq_ex else poll_cq
@@ -1178,6 +1315,9 @@ def xrc_traffic(client, server, is_cq_ex=False, send_op=None):
     send_element_idx = 1 if send_op else 0
     for _ in range(client.num_msgs):
         for i in range(server.qp_count):
+            if force_page_faults:
+                madvise(client.mr.buf, client.msg_size)
+                madvise(server.mr.buf, server.msg_size)
             c_send_wr = get_send_elements(client, False)[send_element_idx]
             if send_op is None:
                 c_send_wr.set_qp_type_xrc(client.remote_srqn)
@@ -1261,6 +1401,13 @@ def odp_implicit_supported(ctx):
         raise unittest.SkipTest('ODP implicit is not supported')
 
 
+def get_pci_name(dev_name):
+    pci_name = glob.glob(f'/sys/bus/pci/devices/*/infiniband/{dev_name}')
+    if not pci_name:
+        raise unittest.SkipTest(f'Could not find the PCI device of {dev_name}')
+    return pci_name[0].split('/')[5]
+
+
 def requires_eswitch_on(func):
     def inner(instance):
         if not (is_eth(d.Context(name=instance.dev_name), instance.ib_port)
@@ -1271,10 +1418,7 @@ def requires_eswitch_on(func):
 
 
 def eswitch_mode_check(dev_name):
-    pci_name = glob.glob(f'/sys/bus/pci/devices/*/infiniband/{dev_name}')
-    if not pci_name:
-        raise unittest.SkipTest(f'Could not find the PCI device of {dev_name}')
-    pci_name = pci_name[0].split('/')[5]
+    pci_name = get_pci_name(dev_name)
     eswicth_off_msg = f'Device {dev_name} must be in switchdev mode'
     try:
         cmd_out = subprocess.check_output(['devlink', 'dev', 'eswitch', 'show', f'pci/{pci_name}'],
@@ -1286,6 +1430,46 @@ def eswitch_mode_check(dev_name):
     return True
 
 
+def requires_roce_disabled(func):
+    def inner(instance):
+        if is_roce_enabled(instance.dev_name):
+            raise unittest.SkipTest('ROCE must be disabled')
+        return func(instance)
+    return inner
+
+
+def is_roce_enabled(dev_name):
+    pci_name = get_pci_name(dev_name)
+    cmd_out = subprocess.check_output(['devlink', 'dev', 'param', 'show', f'pci/{pci_name}',
+                                       'name', 'enable_roce'],
+                                      stderr=subprocess.DEVNULL)
+    if 'value true' in str(cmd_out):
+        return True
+    return False
+
+
+def requires_encap_disabled_if_eswitch_on(func):
+    def inner(instance):
+        if not (is_eth(d.Context(name=instance.dev_name), instance.ib_port)
+                and encap_mode_check(instance.dev_name)):
+            raise unittest.SkipTest('Encap must be disabled when Eswitch on')
+        return func(instance)
+    return inner
+
+
+def encap_mode_check(dev_name):
+    pci_name = get_pci_name(dev_name)
+    encap_enable_msg = f'Device {dev_name}: Encap must be disabled over switchdev mode'
+    try:
+        cmd_out = subprocess.check_output(['devlink', 'dev', 'eswitch', 'show', f'pci/{pci_name}'],
+                                          stderr=subprocess.DEVNULL)
+        if 'switchdev' in str(cmd_out):
+            if any([i for i in ['encap enable', 'encap-mode basic'] if i in str(cmd_out)]):
+                raise unittest.SkipTest(encap_enable_msg)
+    except subprocess.CalledProcessError:
+        raise unittest.SkipTest(encap_enable_msg)
+    return True
+
 def requires_huge_pages():
     def outer(func):
         def inner(instance):
@@ -1389,7 +1573,7 @@ def post_sq_state_bad_flow(test_obj):
     ah = get_global_ah(test_obj.client, test_obj.gid_index, test_obj.ib_port)
     _, sg = get_send_elements(test_obj.client, False)
     with test_obj.assertRaises(PyverbsRDMAError) as ex:
-        send(test_obj.client, sg, e.IBV_QP_EX_WITH_SEND, new_send=True,
+        send(test_obj.client, sg, e.IBV_WR_SEND, new_send=True,
              qp_idx=qp_idx, ah=ah)
     test_obj.assertEqual(ex.exception.error_code, errno.EINVAL)
 
